/*********************************************************************
 *
 * File:         application.c
 *
 * Description:  NPAC Conferencing System API functions for applications 
 *               management 
 *
 * Author:       Tomasz Stachowiak, NPAC
 *
 * Date:         10-22-96
 *
 *********************************************************************/
#include <stdio.h>
#include <winsock.h>

#include "common.h"
#include "rtp.h"
#include "multicast.h"
#include "database.h"

FILE *alog;

void ncsReadAppMessage(ncsApplication *app);

int multicast = 0;
int database = 0;

//extern int errno;


/************************************************************************** 
 *
 * Send request to the Conference Engine 
 *
 **************************************************************************/
int appEngineRequest(ncsApplication *app, short tag, char *data, int len)
{
  MsgHeader header;
  DWORD user_len = 20;

  if(!GetUserName(header.sender.user, &user_len))
	  strcpy(header.sender.user, "unknown");
  header.tag = (unsigned char)tag;
  header.sender.host = gethostid();
  header.dataLength = len;

  if(app->ctrl_socket > 0)
    sendControlMessage(app->ctrl_socket, &header, data, len);

  return 1;
}

/************************************************************************** 
 *
 * Conference Engine messages callbacks
 *
 **************************************************************************/
void appAddMember(ncsApplication *app, char *data)
{
  Participant *newmember;

  newmember = (Participant *)data;

  PRINTF(alog, "Application: member added %s@%d handle %d\n", newmember->user, newmember->host, newmember->handle);

  app->memberActive[newmember->handle] = 1;
  memcpy(&app->member[newmember->handle], newmember, sizeof(Participant));

  app->numMembers++;

  if(app->addMbrCallback != NULL)
    app->addMbrCallback(newmember->handle);

  PRINTF(alog, "App: member added\n");
}

void appDeleteMember(ncsApplication *app, char *data)
{
  int *no;

  no = (int *)data;
  app->memberActive[*no] = 0;
  app->numMembers--;

  if(app->delMbrCallback != NULL)
    app->delMbrCallback(*no);
}

void appPriorityData(ncsApplication *app, char *data, int len)
{
  AppData *appdata;
  int size;

  appdata = (AppData *)data;
  size = len - sizeof(AppData) + 1;
  if(app->dataCallback != NULL)
    app->dataCallback(appdata->tag, appdata->handle, appdata->data, size);
}

void appQuit(ncsApplication *app)
{
  if((multicast) && (app->localHandle >= 0))
    closeMulticast(app->socket);
  closesocket(app->socket);
  closesocket(app->ctrl_socket);
  free(app);
}

/***********************************************************************
 * 
 * Control messages processing procedure
 *
 ***********************************************************************/
void appProcessMessage(ncsApplication *app, unsigned char sessionID, 
		       unsigned char tag, void *data, int len)
{
  InviteData *info;

  info = (InviteData *)data;
  switch((LocalTag)tag)
    {
    case indLocalSessionAddMember:
      PRINTF(alog, "Application: add member\n");
      appAddMember(app, (char *)data);
      break;

    case indLocalSessionDeleteMember:
      PRINTF(alog, "Application: delete member\n");
      appDeleteMember(app, (char *)data);
      break;

    case indLocalSessionLocalMember:
      PRINTF(alog, "Application: local member %d\n", *(int *)data);
      app->localHandle = *(int *)data;
      app->sessionID = sessionID;
      if(multicast)
	initMulticast(app->socket, sessionID, app->port, &app->mcast_addr);

      if(app->startingCallback != NULL)
	app->startingCallback();
      break;

    case indAppData:
      PRINTF(alog, "Application: priority data\n");
      appPriorityData(app, (char *)data, len);
      break;

    case indLocalAppQuit:
      PRINTF(alog, "Application: quit\n");
      if(app->quitCallback != NULL)
	  {
		  PRINTF(alog, "Application: callback quit\n");
		  app->quitCallback();
	  }
      else
		{
			PRINTF(alog, "Application: standard quit\n");
			appQuit(app);
			exit(0);
		}
      break;

    case indDbStart:
      PRINTF(alog, "Application: database start\n");
      dbInit(app->port + 2);
      database = 1;
      break;

    case indDbStop:
      PRINTF(alog, "Application: database start\n");
      database = 0;
      dbClose();
      break;

    default:
      PRINTF(alog, "Application: wrong message tag %d\n", (LocalTag)tag);
    }
}

/***********************************************************************
 * 
 * Control messages reading procedure
 *
 ***********************************************************************/
void ncsReadAppMessage(ncsApplication *app)
{
  char *msg;
  MsgHeader header;
  int err;

  err = readMsgHeader(app->ctrl_socket, &header);
  if(err < 0)
    {
      PRINTF(stderr, "App: data reading error\n");
      return;
    }
  if(err == 0)
    {
      PRINTF(stderr, "App: engine killed, exitting...\n");
      if(app->quitCallback != NULL)
	app->quitCallback();
      else
	{
	  appQuit(app);
	  exit(0);
	}
    }
  if(header.dataLength > 0)
    {
      msg = (char *)malloc(header.dataLength);
      err = readAll(app->ctrl_socket, msg, header.dataLength);
      if(err < 0)
	{
	  PRINTF(stderr, "App: data reading error\n");
	  return;
	}
      if(err == 0)
	{
	  PRINTF(stderr, "App: engine killed, exitting...\n");
	  if(app->quitCallback != NULL)
	    app->quitCallback();
	  else
	    {
	      appQuit(app);
	      exit(0);
	    }
	}
    }
  else
    msg = NULL;
  
  appProcessMessage(app, header.session, header.tag, msg, header.dataLength);
  
  if(msg != NULL)
    free(msg);
}

/************************************************************************** 
 *
 * API functions 
 *
 **************************************************************************/
ncsApplication *ncsAppAttach(int port)
{
  int i;
  ncsApplication *app;
  long hostid;
  WSADATA wsaData;	

#ifndef QUIET
  alog = fopen("app.log", "w");
#endif

  if(WSAStartup(257, &wsaData) != 0)
	  PRINTF(stderr, "Application: can't startup winsock\n");

  app = (ncsApplication *)malloc(sizeof(ncsApplication));
  
  app->port = port;
  app->localHandle = -1;
  
  /* Initialize application data socket */
  if ((app->socket = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
    {
      PRINTF(alog, "Application: can't open socket\n");
      free(app);
      return NULL;
    }
  
  app->addr.sin_family = AF_INET;
  app->addr.sin_addr.s_addr = htonl(INADDR_ANY);
  app->addr.sin_port        = htons(port);
  
  if(bind(app->socket, (struct sockaddr *)&app->addr, sizeof(struct sockaddr)) < 0)
    {
      PRINTF(alog, "Application: can't bind socket\n");
      free(app);
      return NULL;
    }


  /* Initialize application control socket */
  if ((app->ctrl_socket = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
      PRINTF(alog, "Application: can't open control socket\n");
      free(app);
      return NULL;
    }

  /* Set engine address */
  app->daemon_addr.sin_family = AF_INET;
  hostid = gethostid();
  app->daemon_addr.sin_addr.s_addr = hostid;
  app->daemon_addr.sin_port        = htons(DAEMON_PORT);

  if(connect(app->ctrl_socket, (struct sockaddr *)&app->daemon_addr, sizeof(app->daemon_addr)) < 0)
    { 
      PRINTF(alog, "Application: can not connect to engine, exiting...\n");
      exit(0);
    }
  
  
  /* Set app address */
  app->ctrl_addr.sin_family = AF_INET;
  hostid = gethostid();
  app->ctrl_addr.sin_addr.s_addr = hostid;
  app->ctrl_addr.sin_port        = htons(port);

  /* Send request to engine */
  appEngineRequest(app, reqNewLocalApplicationAttach, 
		   (char *)&app->ctrl_addr.sin_port, sizeof(short));

                        
  app->numMembers = 0;
  app->startingCallback = NULL;
  app->quitCallback = NULL;
  app->addMbrCallback = NULL;
  app->delMbrCallback = NULL;

  for(i=0; i<MAX_MEMBERS; i++)
	app->memberActive[i] = 0;

  return app;
}

void ncsAppDettach(ncsApplication *app)
{
  if(app != NULL)
    {
      /* Send request to engine */
      appEngineRequest(app, reqAppDettach, 
		       (char *)&app->port, sizeof(int));
      appQuit(app);

      if(database)
		{
			dbClose();
		}
    }
  
#ifndef QUIET
  fclose(alog);
#endif

  WSACleanup();
}

int ncsAppSendData(ncsApplication *app, int mbr, int type, void *data, int len)
{
  char *ncsdata;
  char *tmp;
  int size;
  int i;
  int sent = 0;
  RTPheader header;

  size = len + RTP_HEADER_SIZE;
  ncsdata = (char *)malloc(size);
  RTPfillHeader(&header, type);
  tmp = ncsdata + RTP_HEADER_SIZE;
  memcpy(tmp, data, len);
  header.ssrc = app->localHandle;
  fillRTPheader(&header, ncsdata);

  if(mbr < 0)  /* send to all members */
  {
	if(multicast)
	{
		if(sendto(app->socket, (char *)ncsdata, size, 
			0, (struct sockaddr *)&app->mcast_addr, sizeof(struct sockaddr_in)) < 0)
		return 0;  
	}
    else
	{
		for(i=0; i<MAX_MEMBERS; i++)
			if(app->memberActive[i])
			{
				if(i != app->localHandle)
				{
					PRINTF(alog, "Application: data sent to user %d, size %d\n", i, size);
					sendPacket(app->socket, app->member[i].host, app->port, (char *)ncsdata, size);
				}
			}
	}

  }
  else   /* send to one specific member */
  {
	if(app->memberActive[mbr])
		if(mbr != app->localHandle)
		{
			sendPacket(app->socket, app->member[mbr].host, app->port, (char *)ncsdata, size);
		}

	/* if message to initiator find initiator */
    if((mbr == 0) && (app->memberActive[mbr] == 0))
	{
		i = 0;
		while((i < MAX_MEMBERS) && (!sent))
	    {
			if(i != app->localHandle)
			{
				PRINTF(alog, "Application: data sent to user %d, size %d\n",i, size);
				sendPacket(app->socket, app->member[i].host, app->port, (char *)ncsdata, size);
				sent = 1;
			}
			i++;
	    }
	}
  }      

  free(ncsdata);

  return 1;
}

int ncsAppSendPriorityData(ncsApplication *app, int mbr, int type, void *data, int len)
{
  AppData *appdata;
  int size;

  size = sizeof(AppData) + len - 1;
  appdata = (AppData *)malloc(size);
  appdata->handle = app->localHandle;
  appdata->recID = mbr;
  appdata->port = app->port;
  appdata->tag = type;
  memcpy(appdata->data, data, len);

  appEngineRequest(app, reqAppData, (char *)appdata, size);

  free(appdata);

  return 1;
}

/*************************************************************************
 * 
 * Proccesses the NCS message incoming from the application control  
 * socket (from the Conference Engine) 
 *
 *************************************************************************/ 
int ncsAppProcessIncoming(ncsApplication *app, fd_set *set)
{
  if(FD_ISSET(app->ctrl_socket, set))
    {
      ncsReadAppMessage(app);
    }
  else
    return 0;

  return 1;
}

/*************************************************************************
 * 
 * Receive the application data(if any) from the application data  
 * socket (from remote applications) 
 *
 *************************************************************************/ 
int ncsAppGetData(ncsApplication *app, fd_set *set, char **data)
{
  int n, remote_addr_len;
  struct sockaddr_in remote_addr;

  remote_addr_len = sizeof(struct sockaddr_in);
  if(FD_ISSET(app->socket, set))
    {
      PRINTF(alog, "Application: there are data\n");

      *data = (char *)malloc(MAX_UDP_SIZE);
      n = recvfrom(app->socket, (char *)*data, MAX_UDP_SIZE, 
		   0, (struct sockaddr *)&remote_addr, &remote_addr_len); 
      if (n < 0)
	{
	  PRINTF(alog, "Application: recvfrom error\n");
	  return 0;
	}

      if(database)
	{
	  dbAdd(*data, n);
	}
    }
  else
    return 0;

  return n;
}  


/***********************************************************************
 * 
 * Get the last file descriptor used by NCS and set the file descriptors
 * set 'set' for all descriptors used by NCS
 *
 ***********************************************************************/
int ncsAppGetFD(ncsApplication *app, fd_set *set)
{
  FD_ZERO(set);
  FD_SET(app->socket, set);
  FD_SET(app->ctrl_socket, set);
  return app->ctrl_socket;
}

int ncsAppGetLocalHandle(ncsApplication *app)
{
  return app->localHandle;
}

int ncsAppGetNumMembers(ncsApplication *app)
{
  return app->numMembers;
}

void ncsAppAddCallback(ncsApplication *app, short tag, void (* function))
{
  switch(tag)
    {
    case tagAppStarting:
      app->startingCallback = (void (*)(void))function;
      break;

    case tagAppQuit:
      app->quitCallback = (void (*)(void))function;
      break;

    case tagAppData:
      app->dataCallback = (void (*)(int, int, void *, int))function;
      break;

    case tagAppAddMbr:
      app->addMbrCallback = (void (*)(int))function;
      break;

    case tagAppDelMbr:
      app->delMbrCallback = (void (*)(int))function;
      break;

    default:
      PRINTF(alog, "Unknown application event tag\n");
    }
}
	
int ncsAppGetPart(ncsApplication *app, int no, char *name, char *host)
{
	if (app->memberActive[no] == 1)
	{
		strcpy(name, app->member[no].user);
		getHostName(app->member[no].host, host);
	}
	else
	{
		strcpy(name, "unknown");
		strcpy(host, "");
		return 0;
	}

	return 1;
}

void readRTPheader(RTPheader *hdr, char *data)
{
	hdr->type = (data[0] & 192) >> 6;
	hdr->p = (data[0] & 32) >> 5;
	hdr->x = (data[0] & 16) >> 4;
	hdr->cc = data[0] & 15;
	hdr->m = (data[1] & 128) >> 7;
	hdr->pt = data[1] & 127;
	hdr->seq = ntohs(*((short *)&data[2]));
	hdr->ts = ntohl(*((int *)&data[4]));
	hdr->ssrc = ntohl(*((int *)&data[8]));
}

void fillRTPheader(RTPheader *hdr, char *data)
{
	short tmp1;
	int tmp2;
	int i;

	for(i=0; i<RTP_HEADER_SIZE; i++)
		data[i] = 0;
	
	data[0] = data[0] | ((hdr->type << 6) & 192);
	data[0] = data[0] | ((hdr->p << 5) & 32);
	data[0] = data[0] | ((hdr->x << 4) & 16);
	data[0] = data[0] | (hdr->cc & 15);
	data[1] = data[1] | ((hdr->m << 7) & 128);
	data[1] = data[1] | (hdr->pt & 127);
	tmp1 = htons(hdr->seq);
	memcpy(&data[2], &tmp1, 2);
	tmp2 = htonl(hdr->ts);
	memcpy(&data[4], &tmp2, 4);
	tmp2 = htonl(hdr->ssrc);
	memcpy(&data[8], &tmp2, 4);
}











